﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

using Hjg.Pngcs;
using Mono.Collections.Generic;
using ProtoBuf;

using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Common.Entities;
using Vintagestory.API.Config;
using Vintagestory.API.Datastructures;
using Vintagestory.API.MathTools;
using Vintagestory.Common;

namespace Automap
{
	public class AutomapSystem
	{
		private Thread cartographer_thread;

		private Snapshotter snapshot;
		private ICoreClientAPI ClientAPI { get; set; }
		private ILogger Logger { get; set; }
		private AChunkRenderer ChunkRenderer { get; set; }
		private JsonGenerator JsonGenerator { get; set; }

		internal const string _mapPath = @"Maps";
		internal const string _chunkPath = @"Chunks";
		internal const uint editThreshold = 9;
		private const string _domain = @"automap";
		private const string chunkFile_filter = @"*_*.png";
		private const string poiFileName = @"poi_binary";
		private const string eoiFileName = @"eoi_binary";
		private const string pointsTsvFileName = @"points_of_interest.tsv";
		private const string plainMetadataFileName = @"map_metadata.txt";
		private static Regex chunkShardRegex = new Regex(@"(?<X>[\d]+)_(?<Z>[\d]+)\.png", RegexOptions.Singleline);

		private ConcurrentDictionary<Vec2i, ColumnCounter> columnCounters = new ConcurrentDictionary<Vec2i, ColumnCounter>(3, 150);
		private ColumnsMetadata chunkTopMetadata;
		internal PointsOfInterest POIs = new PointsOfInterest();
		internal EntitiesOfInterest EOIs = new EntitiesOfInterest();

		internal Dictionary<int, BlockDesignator> BlockID_Designators { get; private set; }
		internal Dictionary<AssetLocation, EntityDesignator> Entity_Designators { get; private set; }
		internal Dictionary<int, string> RockIdCodes { get; private set; }
		internal Dictionary<int, string> AiryIdCodes { get; private set; }

		internal CommandType CurrentState { get; set; }
		//Run status, Chunks processed, stats, center of map....
		private uint nullChunkCount, nullMapCount, updatedChunksTotal;
		private Vec2i startChunkColumn;

		private readonly int chunkSize;
		private string path;
		private IAsset staticMap;
		private PersistedConfiguration configuration;


		public static string AutomapStatusEventKey = @"AutomapStatus";
		public static string AutomapCommandEventKey = @"AutomapCommand";

		public AutomapSystem(ICoreClientAPI clientAPI, ILogger logger, PersistedConfiguration config)
		{
			this.ClientAPI = clientAPI;
			this.Logger = logger;
			chunkSize = ClientAPI.World.BlockAccessor.ChunkSize;

			configuration = config;
			ClientAPI.Event.LevelFinalize += EngageAutomap;

			this.ChunkRenderer = InstantiateChosenRenderer(config.RendererName);

			//Listen on bus for commands
			ClientAPI.Event.RegisterEventBusListener(CommandListener, 1.0, AutomapSystem.AutomapCommandEventKey);


			if (configuration.Autostart)
			{
				CurrentState = CommandType.Run;
				Logger.Notification("Autostart is Enabled.");
			}

		}


		#region Internals
		private void EngageAutomap()
		{
			path = ClientAPI.GetOrCreateDataPath(_mapPath);
			path = ClientAPI.GetOrCreateDataPath(Path.Combine(path, "World_" + ClientAPI.World.Seed));//Add name of World too...'ServerApi.WorldManager.CurrentWorldName'
			ClientAPI.GetOrCreateDataPath(Path.Combine(path, _chunkPath));
			                          
			JsonGenerator = new JsonGenerator(ClientAPI, Logger, path);

			string mapFilename = Path.Combine(path, "automap.html");
			StreamWriter outputText = new StreamWriter(File.Open(mapFilename, FileMode.Create, FileAccess.Write, FileShare.ReadWrite));

			staticMap = ClientAPI.World.AssetManager.Get(new AssetLocation(_domain, "config/automap.html"));
			outputText.Write(staticMap.ToText());
			outputText.Flush();

			Prefill_POI_Designators();
			startChunkColumn = new Vec2i((ClientAPI.World.Player.Entity.Pos.AsBlockPos.X / chunkSize), (ClientAPI.World.Player.Entity.Pos.AsBlockPos.Z / chunkSize));
			chunkTopMetadata = new ColumnsMetadata(startChunkColumn);
			Logger.Notification("AUTOMAP Start {0}", startChunkColumn);
			Reload_Metadata();

			ClientAPI.Event.ChunkDirty += ChunkAChanging;

			cartographer_thread = new Thread(Cartographer)
			{
				Name = "Cartographer",
				Priority = ThreadPriority.Lowest,
				IsBackground = true
			};

			ClientAPI.Event.RegisterGameTickListener(ThreadDecider, 6000);
		}

		private void ChunkAChanging(Vec3i chunkCoord, IWorldChunk chunk, EnumChunkDirtyReason reason)
		{
		Vec2i topPosition = new Vec2i(chunkCoord.X, chunkCoord.Z);		
		bool newOrEdit = (reason == EnumChunkDirtyReason.NewlyCreated || reason == EnumChunkDirtyReason.NewlyLoaded);
		
		columnCounters.AddOrUpdate(topPosition, 
			                      new ColumnCounter(chunkSize, newOrEdit, chunkCoord), 
			                      (chkPos, chkChng) => chkChng.Update(chunkCoord, chunkSize, newOrEdit)
			                     );
		
		}

		/// <summary>
		/// Cartographer Thread 'decider'
		/// </summary>
		/// <param name="delayed">called delay offset</param>
		private void ThreadDecider(float delayed)
		{

			if (CurrentState == CommandType.Run && (ClientAPI.IsGamePaused != false || ClientAPI.IsShuttingDown != true))
			{
				#if DEBUG
				Logger.VerboseDebug("ThreadDecider re-trigger from [{0}]", cartographer_thread.ThreadState);
				#endif

				if (cartographer_thread.ThreadState.HasFlag(ThreadState.Unstarted))
				{
					cartographer_thread.Start();
				}
				else if (cartographer_thread.ThreadState.HasFlag(ThreadState.WaitSleepJoin))
				{
					//Time to (re)write chunk shards
					cartographer_thread.Interrupt();
				}
				//#if DEBUG
				//ClientAPI.TriggerChatMessage($"Automap {updatedChunksTotal} Updates - MAX (N:{chunkTopMetadata.North_mostChunk},S:{chunkTopMetadata.South_mostChunk},E:{chunkTopMetadata.East_mostChunk}, W:{chunkTopMetadata.West_mostChunk} - TOTAL: {chunkTopMetadata.Count})");
				//#endif
			}
			else if (CurrentState == CommandType.Snapshot)
			{
			//Prepare for taking a snopshot
			if (snapshot == null) {		
				snapshot = new Snapshotter(path, chunkTopMetadata, chunkSize, ClientAPI.World.Seed);
				#if DEBUG
				Logger.VerboseDebug("Starting new Snapshot: {0} Wx{1} Hx{2}", snapshot.fileName, snapshot.Width, snapshot.Height);
				#endif
				snapshot.Take( );
				}
			else if (snapshot != null && snapshot.Finished) {
				#if DEBUG
					Logger.VerboseDebug("COMPLETED Snapshot: {0} Wx{1} Hx{2}, taking {3}", snapshot.fileName, snapshot.Width, snapshot.Height, snapshot.Timer.Elapsed);
				#endif
				snapshot = null;
				CurrentState = CommandType.Run;
				}
			}
		}


		private void Cartographer()
		{
			wake:
			Logger.VerboseDebug("Cartographer thread awoken");

			try
			{
				ColumnCounter ejectedItem ;
				uint updatedChunks = 0;
				uint updatedPixels = 0;

				//-- Should dodge enumerator changing underfoot....at a cost.
				if (!columnCounters.IsEmpty)
				{
					var tempSet = columnCounters.ToArray().Where(cks => cks.Value.WeightedSum > editThreshold) .OrderByDescending(kvp => kvp.Value.WeightedSum);
					UpdateEntityMetadata();

					foreach (var mostActiveCol in tempSet)
					{
						var mapChunk = ClientAPI.World.BlockAccessor.GetMapChunk(mostActiveCol.Key);

						if (mapChunk == null)
						{
							//TODO: REVISIT THIS CHUNK!
							#if DEBUG
							Logger.Warning("SKIP CHUNK: ({0}) - Map Chunk NULL!", mostActiveCol.Key);
							#endif
							nullMapCount++;
							columnCounters.TryRemove(mostActiveCol.Key, out ejectedItem);
							continue;
						}

						ColumnMeta chunkMeta;
						if (chunkTopMetadata.Contains(mostActiveCol.Key))
						{
							chunkMeta = chunkTopMetadata[mostActiveCol.Key];
							#if DEBUG
							Logger.VerboseDebug("Loaded meta-chunk {0}", mostActiveCol.Key);
							#endif
						}
						else
						{
							chunkMeta = CreateColumnMetadata(mostActiveCol, mapChunk);
							#if DEBUG
							Logger.VerboseDebug("Created meta-chunk {0}", mostActiveCol.Key);
							#endif
						}
						ProcessChunkBlocks(mostActiveCol.Key, mapChunk, ref chunkMeta);
						mostActiveCol.Value.SetCutoff(chunkMeta.YMax / chunkSize);

						ChunkRenderer.SetupPngImage(mostActiveCol.Key, path, _chunkPath, ref chunkMeta);
						ChunkRenderer.GenerateChunkPngShard(mostActiveCol.Key, mapChunk, chunkMeta, ref chunkTopMetadata, out updatedPixels);

						if (updatedPixels > 0)
						{
							#if DEBUG
							Logger.VerboseDebug("Wrote top-chunk shard: ({0}) - Weight:{1}, Pixels#:{2}", mostActiveCol.Key, mostActiveCol.Value, updatedPixels);
							#endif
							updatedChunks++;
							chunkTopMetadata.Update(chunkMeta);
							columnCounters.TryRemove(mostActiveCol.Key, out ejectedItem);
						}
						else
						{
							columnCounters.TryRemove(mostActiveCol.Key, out ejectedItem);
							#if DEBUG
							Logger.VerboseDebug("Un-painted chunk shard: ({0}) ", mostActiveCol.Key);
							#endif
						}
					}
				}

				UpdateStatus(this.updatedChunksTotal, this.nullChunkCount, updatedChunks);

				if (updatedChunks > 0)
				{
					//What about chunk updates themselves; a update bitmap isn't kept...
					updatedChunksTotal += updatedChunks;
					JsonGenerator.GenerateJSONMetadata(chunkTopMetadata, startChunkColumn, POIs, EOIs, RockIdCodes);
					updatedChunks = 0;

					//Cleanup in-memory Metadata...
					chunkTopMetadata.ClearMetadata( );
				}

				#if DEBUG
				Logger.VerboseDebug("Clearing Column Counters of: {0} non-written shards", columnCounters.Count);
				#endif

				columnCounters.Clear( );

				//Then sleep until interupted again, and repeat
#if DEBUG
				Logger.VerboseDebug("Thread '{0}' about to sleep indefinitely.", Thread.CurrentThread.Name);
#endif
				Thread.Sleep(Timeout.Infinite);

			}
			catch (ThreadInterruptedException)
			{

#if DEBUG
				Logger.VerboseDebug("Thread '{0}' interupted [awoken]", Thread.CurrentThread.Name);
#endif
				goto wake;

			}
			catch (ThreadAbortException)
			{
#if DEBUG
				Logger.VerboseDebug("Thread '{0}' aborted.", Thread.CurrentThread.Name);
#endif
			}
			finally
			{
#if DEBUG
				Logger.VerboseDebug("Thread '{0}' executing finally block.", Thread.CurrentThread.Name);
#endif
				PersistPointsData();
				Write_PlainMetadata( );
			}
		}

		private void Snap()
		{
			snapshotTake:
#if DEBUG
			Logger.VerboseDebug("Snapshot started");
#endif
			try
			{
				snapshot.Take();
#if DEBUG
				Logger.VerboseDebug("Snapshot sleeping");
#endif
				CurrentState = CommandType.Run;
				Thread.Sleep(Timeout.Infinite);
			}
			catch (ThreadInterruptedException)
			{
#if DEBUG
				Logger.VerboseDebug("Snapshot intertupted");
#endif
				goto snapshotTake;
			}
		}

		private void UpdateStatus(uint totalUpdates, uint voidChunks, uint delta)
		{
			StatusData updateData = new StatusData(totalUpdates, voidChunks, delta, CommandType.Run);

			this.ClientAPI.Event.PushEvent(AutomapStatusEventKey, updateData);
		}

		private void Prefill_POI_Designators()
		{

			this.BlockID_Designators = new Dictionary<int, BlockDesignator>();
			this.Entity_Designators = new Dictionary<AssetLocation, EntityDesignator>();
			this.RockIdCodes = Helpers.ArbitrarytBlockIdHunter(ClientAPI, new AssetLocation(GlobalConstants.DefaultDomain, "rock-"), EnumBlockMaterial.Stone);

			var airBlocksQuery = from airyBlock in ClientAPI.World.Blocks
							 where airyBlock.MatterState == EnumMatterState.Solid
							 where airyBlock.BlockMaterial == EnumBlockMaterial.Plant || airyBlock.BlockMaterial == EnumBlockMaterial.Leaves 
							 where airyBlock.CollisionBoxes == null || airyBlock.CollisionBoxes.Length == 0 || airyBlock.RainPermeable == true			                 
						 	 	select airyBlock;			
			//^^ 'Solid' phase - 'Plant' Blocks without any bounding-box; OR 'Invisible' shapes...
			var invisibleBlocksQuery = from novisBlock in ClientAPI.World.Blocks									   
									   where novisBlock.Shape == null || novisBlock.Shape.Base.EndsWith(GlobalConstants.DefaultDomain, @"invisible")   //Whaat! [ base: "block/basic/invisible" ]
									   	select novisBlock;			
			this.AiryIdCodes = airBlocksQuery.Union(invisibleBlocksQuery).ToDictionary(aBlk => aBlk.BlockId, aBlk => aBlk.Code.Path);

			#if DEBUG
			foreach (var fluffBlock in AiryIdCodes) {
			Logger.VerboseDebug("ID#\t{0}:\t{1} IGNORED", fluffBlock.Key, fluffBlock.Value);
			}
			Logger.VerboseDebug("Ignoring {0} blocks", AiryIdCodes.Count);
			#endif

		//Add special marker types for BlockID's of "Interest", overwrite colour, and method
		Reload_POI_Designators();
		}

		private void Reload_POI_Designators()
		{
		uint poisSetup =0, eoiSetup = 0;
			foreach (var designator in configuration.BlockDesignators)
			{
				if (designator.Enabled == false) continue;
				var blockIDs = Helpers.ArbitrarytBlockIdHunter(ClientAPI, designator.Pattern, designator.Material);
				if (blockIDs.Count > 0) { Logger.VerboseDebug("Designator {0} has {1} associated blockIDs", designator.ToString(), blockIDs.Count); }
				foreach (var entry in blockIDs)
				{
					BlockID_Designators.Add(entry.Key, designator);
					poisSetup++;
				}
			}
			this.ChunkRenderer.BlockID_Designators = BlockID_Designators;
			Logger.VerboseDebug("Connected {0} IDs from {1} Block-Designators", poisSetup, configuration.BlockDesignators.Count );


			foreach (var designator in configuration.EntityDesignators)
			{
				if (designator.Enabled == false) continue;
				//Get Variants first, from EntityTypes...better be populated!
				var matched = ClientAPI.World.EntityTypes.FindAll(entp => entp.Code.BeginsWith(designator.Pattern.Domain, designator.Pattern.Path));

				foreach (var match in matched)
				{					
					Logger.VerboseDebug("Linked Entity: {0} Designator: {1}", match.Code, designator);
					this.Entity_Designators.Add(match.Code, designator);
					eoiSetup++;
				}
			}
			Logger.VerboseDebug("Connected {0} IDs from {1} Entity-Designators", eoiSetup, configuration.EntityDesignators.Count);

		}



		/// <summary>
		/// Store Points/Entity of Interest
		/// </summary>
		private void PersistPointsData()
		{
			//POI and EOI raw dump files ~ WRITE em!
			//var poiRawFile = File.
			string poiPath = Path.Combine(path, poiFileName);
			string eoiPath = Path.Combine(path, eoiFileName);

			if (this.POIs.Count > 0)
			{
				using (var poiFile = File.Open(poiPath, FileMode.Create, FileAccess.Write, FileShare.None))
				{
					Serializer.Serialize<PointsOfInterest>(poiFile, this.POIs);
					poiFile.Flush(true);
				}
			}

			if (this.EOIs.Count > 0)
			{
				using (var eoiFile = File.Open(eoiPath, FileMode.Create, FileAccess.Write, FileShare.None))
				{
					Serializer.Serialize<EntitiesOfInterest>(eoiFile, this.EOIs);
					eoiFile.Flush(true);
				}
			}

			//Create Easy to Parse TSV file for tool/human use....
			string pointsTsvPath = Path.Combine(path, pointsTsvFileName);

			using (var tsvWriter = new StreamWriter(pointsTsvPath, false, Encoding.UTF8))
			{
				tsvWriter.WriteLine("Name\tDescription\tLocation\tTime\tDestination\tEntity_UID");
				foreach (var point in this.POIs)
				{
					tsvWriter.Write(point.Name + "\t");
					var notes = point.Notes
						.Replace('\n', '\x001f')
						.Replace("\t", "\\t")
						.Replace("\\", "\\\\");
					tsvWriter.Write(notes + "\t");
					tsvWriter.Write(point.Location.PrettyCoords(ClientAPI) + "\t");
					tsvWriter.Write(point.Timestamp.ToString("u") + "\t");
					tsvWriter.Write((point.Destination != null ? point.Destination.PrettyCoords(ClientAPI) : "---") +"\t");
					tsvWriter.Write("null\t");
					tsvWriter.WriteLine();
				}
				foreach (var entity in this.EOIs)
				{
					tsvWriter.Write(entity.Name + "\t");
					var notes = entity.Notes
						.Replace('\n', '\x001f')
						.Replace("\t", "\\t")
						.Replace("\\", "\\\\");
					tsvWriter.Write(notes + "\t");
					tsvWriter.Write(entity.Location.PrettyCoords(ClientAPI) + "\t");
					tsvWriter.Write(entity.Timestamp.ToString("u") + "\t");
					tsvWriter.Write("---\t");
					tsvWriter.Write(entity.EntityId.ToString("D"));
					tsvWriter.WriteLine();
				}
				tsvWriter.WriteLine();
				tsvWriter.Flush();
			}

		}

		private void Write_PlainMetadata( )
		{ 
		string metaPath = Path.Combine(path, plainMetadataFileName);

		using (var metaDataFile = File.Open(metaPath,FileMode.Create)) {
		using (var mdWriter = new StreamWriter(metaDataFile, Encoding.ASCII)) 
			{
				mdWriter.WriteLine("WorldSeed {0}", ClientAPI.World.Seed);
				mdWriter.WriteLine("PlayerChunkCoords {0:D} {1:D}", startChunkColumn.X, startChunkColumn.Y);
				mdWriter.WriteLine("DefaultSpawnPos {0:D} {1:D} {2:D}", ClientAPI.World.DefaultSpawnPosition.AsBlockPos.X,ClientAPI.World.DefaultSpawnPosition.AsBlockPos.Y,ClientAPI.World.DefaultSpawnPosition.AsBlockPos.Z);
				//mdWriter.WriteLine("CurrentPlayerSpawn", ClientAPI.World.Player.WorldData.EntityPlayer.);
				mdWriter.WriteLine("ChunkSize {0}", chunkSize);
				mdWriter.WriteLine("SeaLevel {0:D}", ClientAPI.World.SeaLevel);
				mdWriter.WriteLine("WorldSize {0:D} {1:D} {2:D}", ClientAPI.World.BulkBlockAccessor.MapSizeX, ClientAPI.World.BulkBlockAccessor.MapSizeY,ClientAPI.World.BulkBlockAccessor.MapSizeZ);
				mdWriter.WriteLine("RegionSize {0:D}", ClientAPI.World.BulkBlockAccessor.RegionSize);
				mdWriter.WriteLine("AMVersion '{0}'", ClientAPI.Self().Info.Version);
				mdWriter.WriteLine("PlayTime {0:F1}", ClientAPI.InWorldEllapsedMilliseconds / 1000);
				mdWriter.WriteLine("GameDate {0}", ClientAPI.World.Calendar.PrettyDate());
				mdWriter.WriteLine("Chunks {0:D}", chunkTopMetadata.Count);
				mdWriter.WriteLine("Chunks Updated {0:D}", updatedChunksTotal);
				mdWriter.WriteLine("Null Chunks {0:D}", nullChunkCount);	
				mdWriter.Flush( );
			}
		}
		}
			

		private ColumnMeta CreateColumnMetadata(KeyValuePair<Vec2i, ColumnCounter> mostActiveCol, IMapChunk mapChunk)
		{
			ColumnMeta data = new ColumnMeta(mostActiveCol.Key.Copy(), ClientAPI, (byte) chunkSize, (ClientAPI.World.BlockAccessor.MapSizeY / chunkSize));
			BlockPos equivBP = new BlockPos(mostActiveCol.Key.X * chunkSize,
											mapChunk.YMax,
											mostActiveCol.Key.Y * chunkSize);

			var climate = ClientAPI.World.BlockAccessor.GetClimateAt(equivBP);
			data.UpdateFieldsFrom(climate, mapChunk, TimeSpan.FromHours(ClientAPI.World.Calendar.TotalHours));

			return data;
		}

		/// <summary>
		/// Reload chunk bounds from chunk shards
		/// </summary>
		/// <returns>The metadata.</returns>
		private void Reload_Metadata()
		{
			var shardsDir = new DirectoryInfo( Path.Combine(path, _chunkPath) );

			if (!shardsDir.Exists)
			{
				#if DEBUG
				Logger.VerboseDebug("Could not open world map (shards) directory");
				#endif
				return;
			}
			var shardFiles = shardsDir.GetFiles(chunkFile_filter);

			if (shardFiles.Length > 0)
			{
				#if DEBUG
				Logger.VerboseDebug("Metadata reloading from {0} shards", shardFiles.Length);
				#endif

				foreach (var shardFile in shardFiles)
				{

					if (shardFile.Length < 1024) continue;
					var result = chunkShardRegex.Match(shardFile.Name);
					if (!result.Success) continue;

					int X_chunk_pos = int.Parse(result.Groups["X"].Value);
					int Z_chunk_pos = int.Parse(result.Groups["Z"].Value);

					try
					{
						using (var fileStream = shardFile.OpenRead())
						{

							PngReader pngRead = new PngReader(fileStream);
							pngRead.ReadSkippingAllRows();
							pngRead.End();
							//Parse PNG chunks for METADATA in shard
							PngMetadataChunk metadataFromPng = pngRead.GetChunksList().GetById1(PngMetadataChunk.ID) as PngMetadataChunk;
							var column = metadataFromPng.ChunkMetadata;
							if (column.PrettyLocation == null)
								column = column.Reload(ClientAPI);
							chunkTopMetadata.Add(column);
						}

					}
					catch (PngjException someEx)
					{
						Logger.Error("PNG Corruption file '{0}' - Reason: {1}", shardFile.Name, someEx);
						continue;
					}
					catch (ProtoException protoEx) 
					{
						Logger.Error("ProtoBuf invalid! file:'{0}' - Reason: {1}", shardFile.Name, protoEx);
						continue;
					}
				}
			}

			//POI and EOI raw dump files ~ reload em!
			//var poiRawFile = File.
			string poiPath = Path.Combine(path, poiFileName);
			string eoiPath = Path.Combine(path, eoiFileName);

			if (File.Exists(poiPath))
			{
				using (var poiFile = File.OpenRead(poiPath))
				{
					this.POIs = Serializer.Deserialize<PointsOfInterest>(poiFile);
					Logger.VerboseDebug("Reloaded {0} POIs from file.", this.POIs.Count);
				}
			}

			if (File.Exists(eoiPath))
			{
				using (var eoiFile = File.OpenRead(eoiPath))
				{
					this.EOIs = Serializer.Deserialize<EntitiesOfInterest>(eoiFile);
					Logger.VerboseDebug("Reloaded {0} EOIs from file.", this.EOIs.Count);
				}
			}

		}



		/// <summary>
		/// Does the heavy lifting of Scanning columns of chunks - scans for BlockEntity, creates Heightmap and stats...
		/// </summary>
		/// <param name="key">Chunk Coordinate</param>
		/// <param name="mapChunk">Map chunk.</param>
		/// <param name="chunkMeta">Chunk metadata</param>
		private void ProcessChunkBlocks(Vec2i key, IMapChunk mapChunk, ref ColumnMeta chunkMeta)
		{
			int targetChunkY = mapChunk.YMax / chunkSize;//Surface ish... 
			byte chunkTally = 0;

		#if DEBUG
		Logger.VerboseDebug("Start col @ X{0} Y{1} Z{2} !", key.X, targetChunkY, key.Y);
		#endif

		chunkMeta.ResetMetadata(ClientAPI.World.BlockAccessor.MapSizeY);

		for (; targetChunkY > 0; targetChunkY--)
			{
				WorldChunk worldChunk = ClientAPI.World.BlockAccessor.GetChunk(key.X, targetChunkY, key.Y) as WorldChunk;

				if (worldChunk == null || worldChunk.BlockEntities == null)
				{
					#if DEBUG
					Logger.VerboseDebug("WORLD chunk: null or empty X{0} Y{1} Z{2} !", key.X, targetChunkY, key.Y);
					#endif
					nullChunkCount++;
					continue;
				}

				if (worldChunk.IsPacked()) 
				{
				#if DEBUG
				Logger.VerboseDebug("WORLD chunk: Compressed: X{0} Y{1} Z{2}", key.X, targetChunkY, key.Y);
				#endif
				worldChunk.Unpack( );//RESEARCH: Thread Unsafe?	
				}

				/*************** Chunk Entities Scanning *********************/
				if (worldChunk.BlockEntities != null && worldChunk.BlockEntities.Count > 0)
				{
					#if DEBUG
					Logger.VerboseDebug("Scan pos.({0}) for BlockEntities# {1}", key, worldChunk.BlockEntities.Count);
					#endif

					foreach (var blockEnt in worldChunk.BlockEntities)
					{
						if (blockEnt.Key != null && blockEnt.Value != null && blockEnt.Value.Block != null && BlockID_Designators.ContainsKey(blockEnt.Value.Block.BlockId))
						{
							var designator = BlockID_Designators[blockEnt.Value.Block.BlockId];
							designator?.SpecialAction(ClientAPI, POIs, blockEnt.Value.Pos.Copy(), blockEnt.Value.Block);
						}
					}
				}

				/********************* Chunk/Column BLOCKs scanning ****************/
				//Heightmap, Stats, block tally

				int X_index, Y_index, Z_index;

				//First Chance fail-safe;
				if (worldChunk.Blocks == null || worldChunk.Blocks.Length <= 0) {
				#if DEBUG
				Logger.VerboseDebug("WORLD chunk; Missing block DATA⁈ X{0} Y{1} Z{2} ⁈", key.X, targetChunkY, key.Y);
				#endif
				nullChunkCount++;
				continue;
				}		

				chunkMeta.ColumnPresense[targetChunkY] = true;
				chunkTally++;
				for (Y_index = 0; Y_index < chunkSize; Y_index++)
				{
					for (Z_index = 0; Z_index < chunkSize; Z_index++)
					{
						for (X_index = 0; X_index < chunkSize; X_index++) 
						{
						var indicie = MapUtil.Index3d(X_index, Y_index, Z_index, chunkSize, chunkSize);

						//'Last' Chance fail-safe;
						if (worldChunk.Blocks == null || worldChunk.Blocks.Length <= 0) {
						#if DEBUG
						Logger.VerboseDebug("Processing Block: Missing block DATA⁈ X{0} Y{1} Z{2} ⁈", X_index, Y_index, Z_index);
						#endif
						nullChunkCount++;
						goto loop_bustout; 
						}

						int aBlockId = worldChunk.Blocks[indicie];

						if (aBlockId == 0 || AiryIdCodes.ContainsKey(aBlockId)) {//Airy blocks,,,
						chunkMeta.AirBlocks++;
						continue;
						}

						if (RockIdCodes.ContainsKey(aBlockId)) {
						if (chunkMeta.RockRatio.ContainsKey(aBlockId))
							chunkMeta.RockRatio[aBlockId]++;
						else
							chunkMeta.RockRatio.Add(aBlockId, 1);
						}

						chunkMeta.NonAirBlocks++;

						ushort localHeight = ( ushort )(Y_index + (targetChunkY * chunkSize));
						//Heightmap - Need to ignore Grass & Snow
						if (localHeight > chunkMeta.HeightMap[X_index, Z_index]) 
							{
							chunkMeta.HeightMap[X_index, Z_index] = localHeight;
							if (localHeight > chunkMeta.YMax) chunkMeta.YMax = localHeight;
							}
						}
					}
				}
				loop_bustout:;
			}
			#if DEBUG
			Logger.VerboseDebug("COLUMN X{0} Z{1}: {2}, processed.", key.X , key.Y, chunkTally + 1);
			#endif
		}

		private void UpdateEntityMetadata()
		{
		#if DEBUG
		Logger.Debug("Presently {0} Entities", ClientAPI.World.LoadedEntities.Count);
		#endif

		var keyList = new long[ClientAPI.World.LoadedEntities.Keys.Count];
		ClientAPI.World.LoadedEntities.Keys.CopyTo(keyList, 0);

	     //'ElementAt'; worse! instead; walk fixed list...
		Entity loadedEntity;
		foreach (var key in keyList)
			{
			if (ClientAPI.World.LoadedEntities.TryGetValue(key, out loadedEntity))
				{		
				#if DEBUG
				//Logger.VerboseDebug($"ENTITY: ({loadedEntity.Value.Code}) = #{loadedEntity.Value.EntityId} {loadedEntity.Value.State} {loadedEntity.Value.LocalPos}    <<<<<<<<<<<<");
				#endif

				var dMatch = Entity_Designators.SingleOrDefault(se => se.Key.Equals(loadedEntity.Code));
				if (dMatch.Value != null) 
					{
					dMatch.Value.SpecialAction(ClientAPI, this.EOIs, loadedEntity.Pos.AsBlockPos.Copy( ), loadedEntity);
					}
				}				
			}
		}

		private void AddNote(string notation)
		{
			var playerNodePoi = new PointOfInterest()
			{
				Name = "Note",
				Location = ClientAPI.World.Player.Entity.Pos.AsBlockPos.Copy(),
				Notes = notation,
				Timestamp = DateTime.UtcNow,
			};

			this.POIs.AddReplace(playerNodePoi);
		}



		private void CommandListener(string eventName, ref EnumHandling handling, IAttribute data)
		{
			//Logger.VerboseDebug("MsgBus RX: AutomapCommandMsg: {0}", data.ToJsonToken());

			CommandData cmdData = data as CommandData;

			switch (cmdData.State)
			{
				case CommandType.Run:
				case CommandType.Stop:
				case CommandType.Snapshot:
					if (CurrentState != cmdData.State)
					{
						CurrentState = cmdData.State;
						ThreadDecider(0.0f);
					}
					break;

				case CommandType.Notation:
					//Add to POI list where player location
					AddNote(cmdData.Notation);
					break;
			}

			ClientAPI.TriggerChatMessage($"Automap commanded to: {cmdData.State} ");

		}
#endregion

		private AChunkRenderer InstantiateChosenRenderer(string rendererName )
		{
		Logger.VerboseDebug("Using '{0}' style Shard Renderer", rendererName);
		switch (rendererName) 
		{				
		case StandardRenderer.Name:
			return new StandardRenderer(ClientAPI, Logger, this.configuration.SeasonalColors);
		
		case AlternateRenderer.Name:
			return new AlternateRenderer(ClientAPI, Logger, this.configuration.SeasonalColors);
	
		case FlatRenderer.Name:
			return new FlatRenderer(ClientAPI, Logger, this.configuration.SeasonalColors);	

		default:
			throw new ArgumentOutOfRangeException("rendererName",rendererName,"That value isn't supported or known...");
		}

		return null;
		}
	}

}
